from django.test import Client
from django.urls import reverse


def test_error_page_direct_access_forbidden(client: Client):
    # error information needs to be in the session to have access
    error_url = reverse("admin-oidc-error")

    response = client.get(error_url)

    assert response.status_code == 403



from urllib.parse import parse_qs, urlparse

from django.contrib.sessions.middleware import SessionMiddleware
from django.http.response import HttpResponseRedirect
from django.test import RequestFactory

import pytest

from mozilla_django_oidc_db.middleware import SessionRefresh
from mozilla_django_oidc_db.models import OpenIDConnectConfig


@pytest.fixture(scope="session")
def dummy_view():
    def get_response(*args, **kwargs):
        pass

    return get_response


@pytest.fixture(scope="session")
def session_middleware(dummy_view):
    return SessionMiddleware(dummy_view)


@pytest.fixture(scope="session")
def session_refresh(dummy_view):
    return SessionRefresh(dummy_view)


@pytest.mark.oidcconfig(enabled=False)
def test_sessionrefresh_oidc_not_enabled(
    dummy_config: OpenIDConnectConfig,
    rf: RequestFactory,
    session_refresh: SessionRefresh,
):
    request = rf.get("/")

    # Running the middleware should return None, since OIDC is disabled
    result = session_refresh(request)

    assert result is None


@pytest.mark.oidcconfig(
    enabled=True,
    oidc_rp_client_id="initial-client-id",
    oidc_rp_scopes_list=["openid", "email"],
)
def test_sessionrefresh_config_always_refreshed(
    dummy_config: OpenIDConnectConfig,
    rf: RequestFactory,
    session_middleware: SessionMiddleware,
    session_refresh: SessionRefresh,
    mocker,
):
    """
    Middleware should refresh the config on every call
    """
    mocker.patch.object(session_refresh, "is_refreshable_url", return_value=True)
    request = rf.get("/")
    session_middleware(request)

    result1 = session_refresh(request)
    assert isinstance(result1, HttpResponseRedirect)
    query1 = parse_qs(urlparse(result1.url).query)
    assert query1["client_id"] == ["initial-client-id"]
    assert query1["scope"] == ["openid email"]

    dummy_config.oidc_rp_client_id = "some-other-id"
    dummy_config.oidc_rp_scopes_list = ["openid", "email", "other_scope"]  # type: ignore
    dummy_config.save()

    result2 = session_refresh(request)
    assert isinstance(result2, HttpResponseRedirect)
    query2 = parse_qs(urlparse(result2.url).query)
    assert query2["client_id"] == ["some-other-id"]
    assert query2["scope"] == ["openid email other_scope"]


@pytest.mark.oidcconfig(enabled=True)
def test_sessionrefresh_config_use_defaults(
    dummy_config,
    settings,
    session_middleware: SessionMiddleware,
    session_refresh: SessionRefresh,
    rf: RequestFactory,
    mocker,
):
    """
    Middleware should respect fallbacks to settings/defaults.
    """
    settings.OIDC_AUTHENTICATION_CALLBACK_URL = "admin:index"
    mocker.patch.object(session_refresh, "is_refreshable_url", return_value=True)

    request = rf.get("/")
    session_middleware(request)

    result = session_refresh(request)

    assert isinstance(result, HttpResponseRedirect)
    query = parse_qs(urlparse(result.url).query)
    assert query["redirect_uri"] == ["http://testserver/admin/"]
    assert len(query["nonce"][0]) == 32  # default set on middleware dynamic_setting


def test_attributeerror_for_non_oidc_attribute(
    dummy_config, session_refresh: SessionRefresh
):
    with pytest.raises(AttributeError):
        session_refresh.__name__  # type: ignore

    # OIDC attributes should never raise AttributeErrors
    assert session_refresh.OIDC_AUTHENTICATION_CALLBACK_URL



from django.db import models
from django.utils.deconstruct import deconstructible
from django.utils.translation import gettext_lazy as _

from django_jsonform.models.fields import ArrayField


@deconstructible
class ClaimFieldDefault:
    """
    Callable default for ClaimField.

    Django's ArrayField requires a callable to be passed for the ``default`` kwarg, to
    avoid sharing a mutable value shared by all instances. This custom class provides
    a straight-forward interface so that defaults can be provided inline rather than
    requiring a function to be defined at the module level, since lambda's cannot be
    serialized for migrations.

    Usage:

    >>> field = ClaimField(default=ClaimFieldDefault("foo", "bar"))
    >>> field.get_default()  # ["foo", "bar"]
    """

    def __init__(self, *bits: str):
        self.bits = list(bits)

    def __eq__(self, other) -> bool:
        if isinstance(other, ClaimFieldDefault):
            return self.bits == other.bits
        return False

    def __call__(self) -> list[str]:
        return self.bits


class ClaimField(ArrayField):
    """
    A field to store a path to claims holding the desired value(s).

    Each item is a segment in the path from the root to leaf for nested claims.
    """

    def __init__(self, *args, **kwargs):
        kwargs["base_field"] = models.CharField(_("claim path segment"), max_length=50)
        super().__init__(*args, **kwargs)



"""
Helpers to work with (dynamic) OIDC configuration.

The utilities here make it easier to work with configuration that lives on a
configuration model instance rather than in Django settings, while also handling
settings that are still defined in the django settings layer.
"""

from typing import Any, Generic, Protocol, TypeVar, overload

from django.apps import apps
from django.core.exceptions import BadRequest
from django.http import HttpRequest

from mozilla_django_oidc.utils import import_from_settings
from typing_extensions import Self, TypedDict, Unpack

from .constants import CONFIG_CLASS_SESSION_KEY
from .models import OpenIDConnectConfigBase


def get_setting_from_config(config: OpenIDConnectConfigBase, attr: str, *args) -> Any:
    """
    Look up a setting from the config record or fall back to Django settings.

    Django settings are defined as ``OIDC_SOME_SETTING``, in upper case, while our
    model fields typically match the name, but in lower case. So, we look up if the
    requested setting exists as an attribut on the configuration instance and use that
    when provided, otherwise we fall back to the django settings module.

    .. note:: A setting may also be defined as a (calculated) property of some kind on
       a/the configuration instance, rather than an explicit model field. That's why
       we use ``hasattr`` checks rather than relying on
       ``config._meta.get_field(some_field)``.
    """
    attr_lowercase = attr.lower()
    if hasattr(config, attr_lowercase):
        # Workaround for OIDC_RP_IDP_SIGN_KEY being an empty string by default.
        # mozilla-django-oidc explicitly checks if `OIDC_RP_IDP_SIGN_KEY` is not `None`
        # https://github.com/mozilla/mozilla-django-oidc/blob/master/mozilla_django_oidc/auth.py#L189
        if (value_from_config := getattr(config, attr_lowercase)) == "":
            return None
        return value_from_config
    return import_from_settings(attr, *args)


T = TypeVar("T")


class SettingsHolder(Protocol[T]):
    def get_settings(self, attr: str, *args: T) -> T: ...


class DynamicSettingKwargs(TypedDict, Generic[T], total=False):
    default: T


class dynamic_setting(Generic[T]):
    """
    Descriptor to lazily access settings while explicitly defining them.

    The instance/class accessing these properties needs to support the ``get_settings``
    method.

    Example usage:

    .. code-block:: python

        class MyBackend(BaseBackend):
            OIDC_OP_TOKEN_ENDPOINT = dynamic_setting[str]()
    """

    _default_set: bool = False
    default: T

    def __init__(self, **kwargs: Unpack[DynamicSettingKwargs[T]]):
        if (default := kwargs.get("default")) is not None:
            self.default = default
            self._default_set = True

    def __set_name__(self, owner: type[object], name: str) -> None:
        self.name = name

    def __repr__(self) -> str:
        default = f" (default: {self.default!r})" if self._default_set else ""
        return f"<dynamic_setting {self.name}{default}>"

    @overload
    def __get__(self, obj: None, objtype: None) -> Self: ...

    @overload
    def __get__(self, obj: SettingsHolder[T], objtype: type[object]) -> T: ...

    def __get__(
        self, obj: SettingsHolder[T] | None, objtype: type[object] | None = None
    ) -> Self | T:
        if obj is None:
            return self
        args = () if not self._default_set else (self.default,)
        return obj.get_settings(self.name, *args)

    def __set__(self, obj: object | None, value: Any) -> None:
        raise AttributeError(f"setting {self.name} is read-only")


def store_config(request: HttpRequest) -> None:
    """
    Store the requested config (class) on the request object.

    mozilla-django-oidc's callback view deletes the state key after it has validated it,
    so our :func:`lookup_config` cannot extract it from the session anymore.
    """
    # Attempt to retrieve the config_class from the session, this only works for users
    # that are actually logged in as Django users
    # The config_class key is added to the state in the OIDCInit.get method.
    # TODO: verify that the state query param is present for error flows! Need to check
    # the OAUTH2 spec for this, but according to ChatGeePeeTee if the request contains
    # it, the callback must have it too.
    config_class = ""
    state_key = request.GET.get("state")
    if state_key and state_key in (states := request.session.get("oidc_states", [])):
        state = states[state_key]
        config_class = state.get("config_class", "")

    if not config_class and (
        _config := request.session.get(CONFIG_CLASS_SESSION_KEY, "")
    ):
        config_class = _config

    try:
        config = apps.get_model(config_class)
    except (LookupError, ValueError) as exc:
        raise BadRequest("Could not look up the referenced config.") from exc

    # Spoofing is not possible since we store it in the server-side session, but there
    # can still be all sorts of programmer mistakes.
    if not issubclass(config, OpenIDConnectConfigBase):
        raise BadRequest("Invalid config referenced.")

    request._oidcdb_config_class = config  # type: ignore


def lookup_config(request: HttpRequest) -> type[OpenIDConnectConfigBase]:
    # cache on request for optimized access
    if (config := getattr(request, "_oidcdb_config_class", None)) is None:
        raise BadRequest("The required config is not available on the request.")
    return config



# Mapping the configuration model fieldnames for endpoints to their
# corresponding names in the OIDC spec
OIDC_MAPPING = {
    "oidc_op_authorization_endpoint": "authorization_endpoint",
    "oidc_op_token_endpoint": "token_endpoint",
    "oidc_op_user_endpoint": "userinfo_endpoint",
    "oidc_op_jwks_endpoint": "jwks_uri",
    "oidc_op_logout_endpoint": "end_session_endpoint",
}

OPEN_ID_CONFIG_PATH = ".well-known/openid-configuration"

CONFIG_CLASS_SESSION_KEY = "_OIDCDB_CONFIG_CLASS"

CLAIM_MAPPING_SCHEMA = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "title": "Claim Mapping",
    "description": "Mapping from user-model fields to OIDC claims",
    "type": "object",
    "properties": {},
    "additionalProperties": {
        "description": "mapping",
        "type": "array",
        "items": {"type": "string"},
    },
}



from typing import Any

import pytest

from mozilla_django_oidc_db.backends import OIDCAuthenticationBackend
from mozilla_django_oidc_db.config import lookup_config
from mozilla_django_oidc_db.middleware import SessionRefresh
from mozilla_django_oidc_db.models import OpenIDConnectConfig
from mozilla_django_oidc_db.views import OIDCAuthenticationRequestView


@pytest.mark.parametrize(
    "setting",
    (
        "OIDC_RP_CLIENT_ID",
        "OIDC_RP_CLIENT_SECRET",
        "OIDC_RP_SIGN_ALGO",
        "OIDC_OP_JWKS_ENDPOINT",
        "OIDC_OP_TOKEN_ENDPOINT",
        "OIDC_OP_USER_ENDPOINT",
        "OIDC_RP_IDP_SIGN_KEY",
    ),
)
def test_backend_without_initialization_request_raises(setting: str):
    backend = OIDCAuthenticationBackend()

    with pytest.raises(AssertionError):
        getattr(backend, setting)


@pytest.mark.parametrize(
    "setting,expected",
    (
        ("OIDC_RP_CLIENT_ID", "testid"),
        ("OIDC_RP_CLIENT_SECRET", "secret"),
        ("OIDC_RP_SIGN_ALGO", "HS256"),
        ("OIDC_OP_JWKS_ENDPOINT", "http://some.endpoint/v1/jwks"),
        ("OIDC_OP_TOKEN_ENDPOINT", "http://some.endpoint/v1/token"),
        ("OIDC_OP_USER_ENDPOINT", "http://some.endpoint/v1/user"),
        ("OIDC_RP_IDP_SIGN_KEY", None),
    ),
)
@pytest.mark.oidcconfig(
    enabled=True,
    oidc_rp_client_id="testid",
    oidc_rp_client_secret="secret",
    oidc_rp_sign_algo="HS256",
    oidc_rp_scopes_list=["openid", "email"],
    oidc_op_jwks_endpoint="http://some.endpoint/v1/jwks",
    oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
    oidc_op_token_endpoint="http://some.endpoint/v1/token",
    oidc_op_user_endpoint="http://some.endpoint/v1/user",
)
def test_backend_reads_settings_from_model(
    dummy_config, callback_request, setting: str, expected: Any
):
    backend = OIDCAuthenticationBackend()
    backend.config_class = lookup_config(callback_request)

    value = getattr(backend, setting)

    assert value == expected


@pytest.mark.oidcconfig(
    enabled=True,
    oidc_rp_client_id="testid",
    oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
)
def test_view_settings_derived_from_model_oidc_enabled(
    dummy_config: OpenIDConnectConfig,
):
    view = OIDCAuthenticationRequestView()

    # verify that the settings are derived from OpenIDConnectConfig
    assert view.OIDC_RP_CLIENT_ID == "testid"
    assert view.OIDC_OP_AUTH_ENDPOINT == "http://some.endpoint/v1/auth"


@pytest.mark.oidcconfig(
    enabled=True,
    oidc_rp_client_id="testid",
    oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
)
def test_middleware_use_falsy_default(
    dummy_config: OpenIDConnectConfig,
):
    middleware = SessionRefresh(lambda x: x)

    # verify that the defaults are allowed
    assert middleware.OIDC_EXEMPT_URLS == []



from typing import Any, ClassVar, Generic, TypeVar, cast

from django.http import HttpRequest

from mozilla_django_oidc.middleware import SessionRefresh as BaseSessionRefresh
from typing_extensions import override

from .config import dynamic_setting, get_setting_from_config
from .models import OpenIDConnectConfig, OpenIDConnectConfigBase

T = TypeVar("T", bound=OpenIDConnectConfigBase)


class BaseRefreshMiddleware(Generic[T], BaseSessionRefresh):
    """
    Point the middleware to a particular config class to use.

    This base class sets up the dynamic settings mechanism.
    """

    _config: T
    config_class: ClassVar[type[OpenIDConnectConfigBase]]
    """
    The config model/class to get the endpoints/credentials from.
    """

    OIDC_EXEMPT_URLS = dynamic_setting[list[str]](default=[])
    OIDC_OP_AUTHORIZATION_ENDPOINT = dynamic_setting[str]()
    OIDC_RP_CLIENT_ID = dynamic_setting[str]()
    OIDC_STATE_SIZE = dynamic_setting[int](default=32)
    OIDC_AUTHENTICATION_CALLBACK_URL = dynamic_setting[str](
        default="oidc_authentication_callback"
    )
    OIDC_RP_SCOPES = dynamic_setting[str](default="openid email")
    OIDC_USE_NONCE = dynamic_setting[bool](default=True)
    OIDC_NONCE_SIZE = dynamic_setting[int](default=32)

    def __init__(self, get_response):
        # `super().__init__` is not called here, because it calls self.get_setting()
        # The retrieval of these settings is handled via dynamic settings above.
        super(BaseSessionRefresh, self).__init__(get_response=get_response)

    @override
    def get_settings(self, attr: str, *args: Any) -> Any:  # type: ignore
        """
        Look up the request setting from the database config.

        We cache the resolved config instance on the middleware instance for when
        settings are repeatedly looked up. Note however, that we also override the
        __call__ method to delete this cached property, as a middleware instance lives
        for the entire lifecycle of the applicatation, and each request must not look
        at stale cached configuration.
        """
        if (config := getattr(self, "_config", None)) is None:
            # django-solo and type checking is challenging, but a new release is on the
            # way and should fix that :fingers_crossed:
            config = cast(T, self.config_class.get_solo())
            self._config = config
        return get_setting_from_config(config, attr, *args)

    def __call__(self, request: HttpRequest):
        # reset the python-level cache for each request
        if hasattr(self, "_config"):
            del self._config
        return super().__call__(request)

    def process_request(self, request):
        # do nothing if the configuration is not enabled
        if not self.get_settings("ENABLED"):
            return None
        return super().process_request(request)


class SessionRefresh(BaseRefreshMiddleware[OpenIDConnectConfig]):
    config_class = OpenIDConnectConfig



from django.contrib.auth.signals import user_logged_in
from django.dispatch import receiver

from .constants import CONFIG_CLASS_SESSION_KEY


@receiver([user_logged_in], dispatch_uid="oidcdb.set_config_class")
def set_oidcdb_config_class_on_session(sender, user, request, **kwargs):
    """
    Record the OIDC config class on the session, this is needed so the callback view
    can retrieve the config in case of a SessionRefresh flow
    """
    if hasattr(user, "_oidcdb_config_class"):
        request.session[CONFIG_CLASS_SESSION_KEY] = user._oidcdb_config_class



from django_setup_configuration.configuration import BaseConfigurationStep
from django_setup_configuration.exceptions import ConfigurationRunFailed

from mozilla_django_oidc_db.forms import OpenIDConnectConfigForm
from mozilla_django_oidc_db.models import OpenIDConnectConfig
from mozilla_django_oidc_db.setup_configuration.models import (
    AdminOIDCConfigurationModel,
    OIDCDiscoveryEndpoint,
)
from mozilla_django_oidc_db.utils import get_groups_by_name


class AdminOIDCConfigurationStep(BaseConfigurationStep[AdminOIDCConfigurationModel]):
    """
    Configure admin login via OpenID Connect
    """

    verbose_name = "Configuration for admin login via OpenID Connect"
    config_model = AdminOIDCConfigurationModel
    namespace = "oidc_db_config_admin_auth"
    enable_setting = "oidc_db_config_enable"

    def execute(self, model: AdminOIDCConfigurationModel) -> None:

        all_settings = {
            "enabled": model.enabled,
            "oidc_rp_client_id": model.oidc_rp_client_id,
            "oidc_rp_client_secret": model.oidc_rp_client_secret,
            "oidc_rp_sign_algo": model.oidc_rp_sign_algo,
            "oidc_rp_scopes_list": model.oidc_rp_scopes_list,
            "oidc_op_jwks_endpoint": model.oidc_op_jwks_endpoint,
            "oidc_token_use_basic_auth": model.oidc_token_use_basic_auth,
            "oidc_rp_idp_sign_key": model.oidc_rp_idp_sign_key,
            "oidc_op_logout_endpoint": model.oidc_op_logout_endpoint,
            "oidc_use_nonce": model.oidc_use_nonce,
            "oidc_nonce_size": model.oidc_nonce_size,
            "oidc_state_size": model.oidc_state_size,
            "oidc_keycloak_idp_hint": model.oidc_keycloak_idp_hint,
            "userinfo_claims_source": model.userinfo_claims_source,
            "username_claim": model.username_claim,
            "claim_mapping": model.claim_mapping,
            "groups_claim": model.groups_claim,
            "sync_groups": model.sync_groups,
            "sync_groups_glob_pattern": model.sync_groups_glob_pattern,
            "make_users_staff": model.make_users_staff,
            "superuser_group_names": model.superuser_group_names,
            "default_groups": get_groups_by_name(
                model.default_groups, model.sync_groups_glob_pattern, model.sync_groups
            ),
        }

        if isinstance(model.endpoint_config, OIDCDiscoveryEndpoint):
            all_settings.update(
                oidc_op_discovery_endpoint=model.endpoint_config.oidc_op_discovery_endpoint,
            )
        else:
            all_settings.update(
                oidc_op_authorization_endpoint=model.endpoint_config.oidc_op_authorization_endpoint,
                oidc_op_token_endpoint=model.endpoint_config.oidc_op_token_endpoint,
                oidc_op_user_endpoint=model.endpoint_config.oidc_op_user_endpoint,
            )

        form = OpenIDConnectConfigForm(
            instance=OpenIDConnectConfig.get_solo(),
            data=all_settings,
        )
        if not form.is_valid():
            raise ConfigurationRunFailed(
                "Admin OIDC configuration field validation failed",
                form.errors.as_json(),
            )
        form.save()



# Generated by Django 3.2.16 on 2022-12-13 12:47

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("mozilla_django_oidc_db", "0010_openidconnectconfig_userinfo_claims_source"),
    ]

    operations = [
        migrations.AlterField(
            model_name="openidconnectconfig",
            name="userinfo_claims_source",
            field=models.CharField(
                choices=[
                    ("userinfo_endpoint", "Userinfo endpoint"),
                    ("id_token", "ID token"),
                ],
                default="userinfo_endpoint",
                help_text="Indicates the source from which the user information claims should be extracted.",
                max_length=100,
                verbose_name="user information claims extracted from",
            ),
        ),
    ]



from __future__ import annotations

import logging
from collections.abc import Collection
from typing import Any, TypeAlias, cast

from django.contrib.auth import get_user_model
from django.contrib.auth.models import (
    AbstractBaseUser,
    AbstractUser,
    AnonymousUser,
    Group,
)
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest

import requests
from glom import Path, glom
from mozilla_django_oidc.auth import OIDCAuthenticationBackend as BaseBackend
from typing_extensions import override

from .config import dynamic_setting, get_setting_from_config, lookup_config
from .exceptions import MissingIdentifierClaim
from .jwt import verify_and_decode_token
from .models import OpenIDConnectConfigBase, UserInformationClaimsSources
from .typing import ClaimPath, JSONObject
from .utils import extract_content_type, get_groups_by_name, obfuscate_claims

logger = logging.getLogger(__name__)

AnyUser: TypeAlias = AnonymousUser | AbstractBaseUser

missing = object()


class OIDCAuthenticationBackend(BaseBackend):
    """
    Custom backend modifying the upstream package behaviour.

    This backend is meant to look up the configuration (class) to use, which is set
    in the session by the init view. It does this by grabbing the state parameter from
    the GET params to look up the state set in the authentication request view. The
    specified configuration parameters are then loaded and applied.

    No configuration is loaded in :meth:`__init__` at all, instead we define properties
    to dynamically look this up. Django instantiates backends *a lot*, e.g. during
    permission checks. We only support the :meth:`authenticate` entrypoint like the
    upstream library.
    """

    request: HttpRequest | None = None  # set during the authenticate call

    config_class: type[OpenIDConnectConfigBase]

    # These should be functionaly equivalent to
    # :class:`mozilla_django_oidc.auth.OIDCAuthenticationBackend`.
    OIDC_OP_TOKEN_ENDPOINT = dynamic_setting[str]()
    OIDC_OP_USER_ENDPOINT = dynamic_setting[str]()
    OIDC_OP_JWKS_ENDPOINT = dynamic_setting[str | None](default=None)
    OIDC_RP_CLIENT_ID = dynamic_setting[str]()
    OIDC_RP_CLIENT_SECRET = dynamic_setting[str]()
    OIDC_RP_SIGN_ALGO = dynamic_setting[str](default="HS256")
    OIDC_RP_IDP_SIGN_KEY = dynamic_setting[str | None](default=None)

    # Custom config/settings from our own models. These map to model fields or
    # properties on the OpenIDConnectConfigBase model. The settings with defaults can
    # be provided by subclasses of OpenIDConnectConfigBase or plain old Django settings.
    OIDCDB_SENSITIVE_CLAIMS = dynamic_setting[Collection[ClaimPath]]()
    OIDCDB_USERNAME_CLAIM = dynamic_setting[ClaimPath]()
    OIDCDB_USERINFO_CLAIMS_SOURCE = dynamic_setting[UserInformationClaimsSources](
        default=UserInformationClaimsSources.userinfo_endpoint
    )
    OIDCDB_USERNAME_CASE_SENSITIVE = dynamic_setting[bool](default=True)
    OIDCDB_CLAIM_MAPPING = dynamic_setting[dict[str, ClaimPath]](default={})
    OIDCDB_GROUPS_CLAIM = dynamic_setting[ClaimPath](default=[])
    OIDCDB_DEFAULT_GROUPS = dynamic_setting[Collection[str]](default=[])
    OIDCDB_SYNC_MISSING_GROUPS = dynamic_setting[bool](default=True)
    OIDCDB_SYNC_GROUPS_GLOB_PATTERN = dynamic_setting[str](default="*")
    OIDCDB_MAKE_USERS_STAFF = dynamic_setting[bool](default=False)
    OIDCDB_SUPERUSER_GROUP_NAMES = dynamic_setting[Collection[str]](default=[])

    @override
    def __init__(self, *args, **kwargs) -> None:
        # Deliberately empty, we discard all initialization from the parent class which
        # requires a config_class to be set. Even if we set it (declaratively), this is
        # not viable because it performs DB/cache IO to look up the config instance,
        # which would happen as well when Django goes through the auth backends for
        # permission checks.
        #
        # See https://github.com/maykinmedia/mozilla-django-oidc-db/issues/30

        # django-stubs returns AbstractBaseUser, but we depend on properties of
        # AbstractUser.
        self.UserModel = cast(AbstractUser, get_user_model())

    @override
    def get_settings(self, attr: str, *args: Any) -> Any:  # type: ignore
        """
        Override the upstream library get_settings.

        Upstream is django-settings based, and we store configuration in database
        records instead. We look up the configuration from the DB and check if the
        requested setting is defined there or not. If not, it is taken from the Django
        settings.
        """
        assert hasattr(self, "config_class"), (
            "The config must be loaded from the authenticate entrypoint. It looks like "
            "you're trying to access configuration before this was called."
        )
        if (config := getattr(self, "_config", None)) is None:
            # django-solo and type checking is challenging, but a new release is on the
            # way and should fix that :fingers_crossed:
            config = cast(OpenIDConnectConfigBase, self.config_class.get_solo())
            self._config = config
            self._validate_settings()
        return get_setting_from_config(config, attr, *args)

    def _validate_settings(self):
        # Equivalent of checks in upstream __init__ method
        if (
            self.OIDC_RP_SIGN_ALGO.startswith("RS")
            or self.OIDC_RP_SIGN_ALGO.startswith("ES")
        ) and (
            self.OIDC_RP_IDP_SIGN_KEY is None and self.OIDC_OP_JWKS_ENDPOINT is None
        ):
            msg = "{} alg requires OIDC_RP_IDP_SIGN_KEY or OIDC_OP_JWKS_ENDPOINT to be configured."
            raise ImproperlyConfigured(msg.format(self.OIDC_RP_SIGN_ALGO))

    def _check_candidate_backend(self) -> bool:
        return self.get_settings("enabled")

    # The method signature is checked by django when selecting a suitable backend. Our
    # signature is more strict than the upstream library. Check the upstream
    # `OIDCAuthenticationCallbackView` for the `auth.authenticate(**kwargs)` call if this
    # needs updating.
    @override
    def authenticate(  # type: ignore
        self,
        request: HttpRequest | None,
        nonce: str | None = None,
        code_verifier: str | None = None,
    ) -> AnyUser | None:
        """
        Authenticate the user with OIDC *iff* the conditions are met.

        Return ``None`` to skip to the next backend, raise
        :class:`django.core.exceptions.PermissionDenied` to stop in our tracks. Return
        a user object (real or anonymous) to signify success.
        """
        # if we don't get a request, we can't check anything, so skip to the next
        # backend. We need to grab the state and code from request.GET for OIDC.
        if request is None:
            return None

        # Load the config to apply and check if this backend should be considered to
        # authenticate the user.
        self.config_class = lookup_config(request)
        is_candidate = self._check_candidate_backend()
        if not is_candidate:
            return None

        # Allright, now try to actually authenticate the user.
        user = super().authenticate(request, nonce=nonce, code_verifier=code_verifier)

        # Store the config class name on the user, so that we can store this in the user's
        # session after they have been successfully authenticated (by listening to the `user_logged_in` signal)
        if user:
            options = self.config_class._meta
            user._oidcdb_config_class = f"{options.app_label}.{options.object_name}"  # type: ignore

        return user

    def _extract_username(
        self, claims: JSONObject, *, raise_on_empty: bool = False
    ) -> str:
        """
        Given the claims and the dynamic config, extract the username value.
        """
        claim_bits = self.OIDCDB_USERNAME_CLAIM
        unique_id = glom(claims, Path(*claim_bits), default="")
        if raise_on_empty and not unique_id:
            raise MissingIdentifierClaim(claim_bits=claim_bits)
        return unique_id

    @override
    def verify_claims(self, claims: JSONObject) -> bool:
        """Verify the provided claims to decide if authentication should be allowed."""
        assert claims, "Empty claims should have been blocked earlier"

        obfuscated_claims = obfuscate_claims(claims, self.OIDCDB_SENSITIVE_CLAIMS)

        logger.debug("OIDC claims received: %s", obfuscated_claims)

        # check if we have an identifier
        try:
            self._extract_username(claims, raise_on_empty=True)
        except MissingIdentifierClaim as exc:
            logger.error(
                "'%s' not in OIDC claims, cannot proceed with authentication",
                " > ".join(exc.claim_bits),
                exc_info=exc,
            )
            return False
        return True

    @override
    def get_userinfo(
        self, access_token: str, id_token: str, payload: JSONObject
    ) -> JSONObject:
        """
        Extract the user information, configurable whether to use the ID token or
        the userinfo endpoint for this
        """
        if self.OIDCDB_USERINFO_CLAIMS_SOURCE == UserInformationClaimsSources.id_token:
            logger.debug("Extracting user information from ID token")
            return payload

        logger.debug("Retrieving user information from userinfo endpoint")

        # copy of upstream get_userinfo which doesn't support application/jwt yet.
        # Overridden to handle application/jwt responses.
        # See https://github.com/mozilla/mozilla-django-oidc/issues/517
        #
        # Specifying the preferred format in the ``Accept`` header does not work with
        # Keycloak, as it depends on the client settings.
        user_response = requests.get(
            self.OIDC_OP_USER_ENDPOINT,
            headers={
                "Authorization": "Bearer {0}".format(access_token),
            },
            verify=self.get_settings("OIDC_VERIFY_SSL", True),
            timeout=self.get_settings("OIDC_TIMEOUT", None),
            proxies=self.get_settings("OIDC_PROXY", None),
        )
        user_response.raise_for_status()

        # From https://openid.net/specs/openid-connect-core-1_0.html#UserInfoResponse
        #
        # > The UserInfo Endpoint MUST return a content-type header to indicate which
        # > format is being returned.
        content_type = extract_content_type(user_response.headers["Content-Type"])
        match content_type:
            case "application/json":
                # the default case of upstream library
                return user_response.json()
            case "application/jwt":
                token = user_response.content
                # get the key from the configured keys endpoint
                # XXX: tested with asymmetric encryption. algorithms like HS256 rely on
                # out-of-band key exchange and are currently not supported until such a
                # case arrives.
                key = self.retrieve_matching_jwk(token)
                payload = verify_and_decode_token(token, key)
                return payload
            case _:
                raise ValueError(
                    f"Got an invalid Content-Type header value ({content_type}) "
                    "according to OpenID Connect Core 1.0 standard. Contact your "
                    "vendor."
                )

    @override
    def filter_users_by_claims(self, claims: JSONObject):
        """Return all users matching the specified subject."""
        username = self._extract_username(claims)
        assert username, (
            "A username must be provided. Did you forget to check empty usernames "
            "in the claim verification?"
        )

        lookup = self.UserModel.USERNAME_FIELD
        if not self.OIDCDB_USERNAME_CASE_SENSITIVE:
            lookup += "__iexact"

        return self.UserModel.objects.filter(**{lookup: username})

    @override
    def create_user(self, claims: JSONObject) -> AnyUser:
        """Return object for a newly created user account."""
        username = self._extract_username(claims)
        logger.debug("Creating OIDC user: %s", username)

        user = self.UserModel.objects.create_user(
            **{self.UserModel.USERNAME_FIELD: username}
        )
        self.update_user(user, claims)
        return user

    @override
    def update_user(self, user: AbstractUser, claims: JSONObject):
        """
        Update existing user with new claims, if necessary save, and return user

        This method checks dynamic settings, which are provided by our
        :class:`~mozilla_django_oidc_db.models.OpenIDConnectConfig` model. If the
        respective fields do not exist on the config model being used, you must make
        sure they exist as Django settings.
        """
        logger.debug("Updating OIDC user %s", user)

        touched_fields: set[str] = set()

        for model_field, claim_bits in self.OIDCDB_CLAIM_MAPPING.items():
            value = glom(claims, Path(*claim_bits), default=missing)
            if value is missing:
                continue
            setattr(user, model_field, value)
            touched_fields.add(model_field)

        # Users can only be promoted to staff. Staff rights are never taken by OIDC.
        if self.OIDCDB_MAKE_USERS_STAFF and not user.is_staff:
            user.is_staff = True
            touched_fields.add("is_staff")

        groups = _get_groups(claims, self.OIDCDB_GROUPS_CLAIM)

        # Assign superuser status to the user if the user is a member of at least one
        # specific group. Superuser status is explicitly removed if the user is not or
        # no longer member of at least one of these groups.
        if self.OIDCDB_SUPERUSER_GROUP_NAMES and groups:
            # superuser is granted if there is overlap between the groups claim and the
            # specified superuser grup names
            make_superuser = bool(set(groups) & set(self.OIDCDB_SUPERUSER_GROUP_NAMES))
            if make_superuser != user.is_superuser:
                user.is_superuser = make_superuser
                touched_fields.add("is_superuser")

        user.save(update_fields=touched_fields)

        _set_user_groups(
            user,
            groups,
            default_group_names=self.OIDCDB_DEFAULT_GROUPS,
            sync_missing_groups=self.OIDCDB_SYNC_MISSING_GROUPS,
            sync_groups_glob=self.OIDCDB_SYNC_GROUPS_GLOB_PATTERN,
        )

        return user


def _get_groups(claims: JSONObject, groups_claim: ClaimPath) -> list[str] | None:
    if not groups_claim:
        return

    # extract the specified claim
    groups: list[str] | str = glom(claims, Path(*groups_claim), default=missing)
    if groups is missing:
        logger.debug(
            "The configured groups claim '%s' was not found in the user info.",
            " > ".join(groups_claim),
        )
        groups = []

    # could be a string instead of a list of strings
    if isinstance(groups, str):
        groups = [groups]

    if not all(isinstance(name, str) for name in groups):
        logger.warning("Aborting! Groups is not a list of strings: %r", groups)
        return

    return groups


def _set_user_groups(
    user: AbstractUser,
    groups: list[str] | None,
    default_group_names: Collection[str],
    sync_missing_groups: bool,
    sync_groups_glob: str,
) -> None:
    """
    Synchronize the groups from the config/claims and the model.
    """
    if groups is None:
        return

    # Incorporate the default groups into the desired state of groups.
    desired_group_names = set(default_group_names) | set(groups)

    # Update the user's group memberships, if required
    django_group_names = set(user.groups.values_list("name", flat=True))
    if set(desired_group_names) == django_group_names:
        return

    # Create missing groups if required
    existing_groups = get_groups_by_name(
        desired_group_names, sync_groups_glob, sync_missing_groups
    )

    # at this point, existing_groups is the full collection of groups that should be
    # set on the user model, because:
    #
    # * desired_group_names is equal to or a superset of the existing groups (superset
    #   if group synchronization is off or groups are excluded by the glob)
    # * the groups currently set on the user may contain groups not present in the
    #   desired group names, as the latter only reflects the default groups + groups in
    #   the claims
    user.groups.set(existing_groups)



import inspect
from collections.abc import Sequence
from typing import Any

from django.apps import AppConfig
from django.conf import settings
from django.core.checks import CheckMessage, Error, Warning, register
from django.utils.module_loading import import_string

from .views import OIDCCallbackView, OIDCInit


def _do_check(
    app_configs: Sequence[AppConfig] | None,
    dotted_path: Any,
    type_error: Error,
    subclass_reference: type,
    subclass_warning: Warning,
) -> list[CheckMessage]:
    if not (
        app_configs is None
        or any(config.name == "mozilla_django_oidc_db" for config in app_configs)
    ):
        return []

    if not isinstance(dotted_path, str):
        return [type_error]

    view_cls = import_string(dotted_path)
    if not inspect.isclass(view_cls) or not issubclass(view_cls, subclass_reference):
        return [subclass_warning]

    return []


@register()
def check_authenticate_class(
    *, app_configs: Sequence[AppConfig] | None, **kwargs
) -> list[CheckMessage]:
    type_error = Error(
        "'settings.OIDC_AUTHENTICATE_CLASS' must be a string that can be imported.",
        hint=(
            "Use 'mozilla_django_oidc_db.views.OIDCAuthenticationRequestView' or a "
            "subclass of 'mozilla_django_oidc_db.views.OIDCInit'."
        ),
        id="mozilla_django_oidc_db.E001",
    )
    subclass_warning = Warning(
        "'settings.OIDC_AUTHENTICATE_CLASS' should be a subclass of 'OIDCInit'.",
        hint=(
            "Use 'mozilla_django_oidc_db.views.OIDCAuthenticationRequestView' or a "
            "subclass of 'mozilla_django_oidc_db.views.OIDCInit'."
        ),
        id="mozilla_django_oidc_db.W001",
    )

    return _do_check(
        app_configs,
        settings.OIDC_AUTHENTICATE_CLASS,
        type_error=type_error,
        subclass_reference=OIDCInit,
        subclass_warning=subclass_warning,
    )


@register()
def check_callback_class(
    *, app_configs: Sequence[AppConfig] | None, **kwargs
) -> list[CheckMessage]:
    type_error = Error(
        "'settings.OIDC_CALLBACK_CLASS' must be a string that can be imported.",
        hint=(
            "Use 'mozilla_django_oidc_db.views.OIDCCallbackView' or a "
            "subclass of it."
        ),
        id="mozilla_django_oidc_db.E002",
    )
    subclass_warning = Warning(
        "'settings.OIDC_CALLBACK_CLASS' should be a subclass of 'OIDCInit'.",
        hint=(
            "Use 'mozilla_django_oidc_db.views.OIDCCallbackView' or a "
            "subclass of it."
        ),
        id="mozilla_django_oidc_db.W002",
    )

    return _do_check(
        app_configs,
        settings.OIDC_CALLBACK_CLASS,
        type_error=type_error,
        subclass_reference=OIDCCallbackView,
        subclass_warning=subclass_warning,
    )



import json

from django import forms
from django.utils.translation import gettext_lazy as _

import requests

from .constants import OIDC_MAPPING, OPEN_ID_CONFIG_PATH
from .models import OpenIDConnectConfig


class OpenIDConnectConfigForm(forms.ModelForm):
    required_endpoints = [
        "oidc_op_authorization_endpoint",
        "oidc_op_token_endpoint",
        "oidc_op_user_endpoint",
    ]
    oidc_mapping = OIDC_MAPPING

    class Meta:
        model = OpenIDConnectConfig
        fields = "__all__"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Only applicable if user has write access
        if self.fields:
            # Required endpoints should be optional in the form, if the can be
            # derived from the discovery endpoint
            for endpoint in self.required_endpoints:
                self.fields[endpoint].required = False

    @classmethod
    def get_endpoints_from_discovery(cls, base_url: str):
        response = requests.get(f"{base_url}{OPEN_ID_CONFIG_PATH}", timeout=10)
        response.raise_for_status()
        configuration = response.json()

        endpoints = {
            model_attr: configuration.get(oidc_attr)
            for model_attr, oidc_attr in cls.oidc_mapping.items()
        }
        return endpoints

    def clean(self):
        cleaned_data = super().clean()

        discovery_endpoint = cleaned_data.get("oidc_op_discovery_endpoint")

        # Derive the endpoints from the discovery endpoint
        if discovery_endpoint:
            try:
                endpoints = self.get_endpoints_from_discovery(discovery_endpoint)
                cleaned_data.update(**endpoints)
            except (
                requests.exceptions.RequestException,
                json.decoder.JSONDecodeError,
            ):
                raise forms.ValidationError(
                    {
                        "oidc_op_discovery_endpoint": _(
                            "Something went wrong while retrieving the configuration."
                        )
                    }
                )
        else:
            # Verify that the required endpoints were derived from the
            # discovery endpoint
            for field in self.required_endpoints:
                if not cleaned_data.get(field):
                    self.add_error(field, _("This field is required."))

        return cleaned_data



import logging

from django.contrib.auth.models import Group, User
from django.core.exceptions import BadRequest, ImproperlyConfigured
from django.http import HttpRequest
from django.test import RequestFactory

import pytest
from mozilla_django_oidc.auth import OIDCAuthenticationBackend as UpstreamBackend

from mozilla_django_oidc_db.backends import OIDCAuthenticationBackend
from mozilla_django_oidc_db.config import lookup_config
from mozilla_django_oidc_db.models import (
    OpenIDConnectConfig,
    UserInformationClaimsSources,
)
from mozilla_django_oidc_db.views import OIDCAuthenticationRequestView
from testapp.backends import MockBackend
from testapp.models import EmptyConfig

from .custom_config import oidc_init_disabled

#
# DYNAMIC CONFIGURATION TESTS
#


@pytest.mark.oidcconfig(enabled=False)
def test_authenticate_oidc_not_enabled(dummy_config, callback_request: HttpRequest):
    backend = OIDCAuthenticationBackend()

    user = backend.authenticate(callback_request)

    # Authentication with the backend should not return a result,
    # because OIDC is not enabled
    assert user is None


@pytest.mark.callback_request(init_view=oidc_init_disabled)
def test_authentication_loads_config_from_init_state(
    dummy_config, callback_request: HttpRequest
):
    assert dummy_config.enabled is True
    backend = OIDCAuthenticationBackend()
    # check that we correctly initialized our OIDC state
    state_key = callback_request.GET["state"]
    state = callback_request.session["oidc_states"][state_key]
    assert state["config_class"] == "mozilla_django_oidc_db.DisabledConfig"

    user = backend.authenticate(callback_request)

    # Authentication with the backend should not return a result,
    # because OIDC is not enabled
    assert user is None


@pytest.mark.parametrize(
    "setting,expected",
    (
        ("OIDCDB_USERNAME_CASE_SENSITIVE", False),
        ("OIDCDB_CLAIM_MAPPING", {"foo": "bar"}),
        ("OIDCDB_GROUPS_CLAIM", ["roles"]),
        ("OIDCDB_DEFAULT_GROUPS", ["one", "two"]),
        ("OIDCDB_SYNC_MISSING_GROUPS", False),
        ("OIDCDB_SYNC_GROUPS_GLOB_PATTERN", "in:*:tricate"),
        ("OIDCDB_MAKE_USERS_STAFF", True),
        ("OIDCDB_SUPERUSER_GROUP_NAMES", ["we are gods", "bow for us"]),
    ),
)
@pytest.mark.django_db
def test_grabs_config_from_django_settings_if_missing_on_model(
    settings, setting, expected
):
    setattr(settings, setting, expected)
    backend = OIDCAuthenticationBackend()
    backend.config_class = EmptyConfig

    value = getattr(backend, setting)

    assert value == expected


@pytest.mark.parametrize("sign_alg", ("RS**", "ES**"))
@pytest.mark.oidcconfig()
def test_settings_still_validated(dummy_config, settings, sign_alg: str):
    """
    Test that the upstream library settings checks are still performed.
    """
    dummy_config.oidc_rp_sign_algo = sign_alg
    dummy_config.oidc_rp_idp_sign_key = ""
    dummy_config.oidc_op_jwks_endpoint = ""
    dummy_config.save()
    backend = OIDCAuthenticationBackend()
    backend.config_class = type(dummy_config)

    with pytest.raises(ImproperlyConfigured):
        backend.OIDC_RP_CLIENT_ID  # the exact setting doesn't matter

    # check that the same error is raised by the upstream backend
    settings.OIDC_RP_SIGN_ALGO = sign_alg
    settings.OIDC_RP_IDP_SIGN_KEY = None
    settings.OIDC_OP_JWKS_ENDPOINT = None
    with pytest.raises(ImproperlyConfigured):
        UpstreamBackend()


#
# LOGGING AND OBFUSCATION TESTS
#


class SensitiveClaimsConfig(OpenIDConnectConfig):
    class Meta:
        proxy = True
        app_label = "mozilla_django_oidc_db"

    sensitive_claims = (
        ["sensitive_claim1"],
        ["parent", "sensitive_claim2"],
    )


@pytest.mark.callback_request(
    init_view=OIDCAuthenticationRequestView.as_view(config_class=SensitiveClaimsConfig)
)
@pytest.mark.oidcconfig(enabled=True, username_claim=["sub"])
def test_obfuscates_sensitive_claims(
    dummy_config, callback_request: HttpRequest, caplog
):
    caplog.set_level(logging.DEBUG, logger="mozilla_django_oidc_db.backends")
    backend = OIDCAuthenticationBackend()
    backend.request = callback_request
    backend.config_class = lookup_config(callback_request)

    claims_ok = backend.verify_claims(
        {
            "sub": "some-unique-id",
            "sensitive_claim1": "obfuscate-me!",
            "parent": {
                "sensitive_claim2": "obfuscate-me!",
                "other": "not-sensitive",
            },
        }
    )

    assert claims_ok

    # check that the emitted log records have the sensitive claims obfuscated
    log_message: str = next(
        message
        for rec in caplog.records
        if (message := rec.message).startswith("OIDC claims received")
    )

    assert "some-unique-id" not in log_message
    assert "obfuscate-me!" not in log_message
    assert "not-sensitive" in log_message


#
# USER CREATION & UPDATING TESTS
#


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
)
def test_create_user_with_default_config(dummy_config, callback_request: HttpRequest):
    assert not User.objects.exists()
    backend = MockBackend(
        claims={
            "sub": "123456",
            "email": "admin@localhost",
            "given_name": "John",
            "family_name": "Doe",
        }
    )

    user = backend.authenticate(request=callback_request)

    assert user is not None
    assert isinstance(user, User)

    # Verify that a user is created with the correct values
    assert user.pk is not None
    assert user.username == "123456"
    assert user.email == "admin@localhost"
    assert user.first_name == "John"
    assert user.last_name == "Doe"


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    username_claim=["sub"],
)
def test_case_insensitive_username_lookups(
    settings, dummy_config, callback_request: HttpRequest
):
    settings.OIDCDB_USERNAME_CASE_SENSITIVE = False
    existing_user = User.objects.create(username="ADMIN")

    backend = MockBackend(
        claims={
            "sub": "admin",
            "email": "admin@localhost",
            "given_name": "John",
            "family_name": "Doe",
        }
    )

    user = backend.authenticate(request=callback_request)

    assert user == existing_user
    assert User.objects.count() == 1


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    username_claim=["user", "username"],
    groups_claim=["user", "groups.names"],
    claim_mapping={
        "email": ["attributes.email"],
        "first_name": ["profile", "given_name"],
        "last_name": ["profile", "family_name"],
    },
)
def test_create_user_with_custom_and_complex_config(
    dummy_config, callback_request: HttpRequest
):
    assert not User.objects.exists()
    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "user": {
                "username": "admin",
                "groups.names": ["supergroup"],
            },
            "profile": {
                "given_name": "John",
                "family_name": "Doe",
            },
            "attributes.email": "admin@localhost",
        }
    )

    user = backend.authenticate(request=callback_request)

    assert user is not None
    assert isinstance(user, User)

    # Verify that a user is created with the correct values
    assert user.pk is not None
    assert user.username == "admin"
    assert user.email == "admin@localhost"
    assert user.first_name == "John"
    assert user.last_name == "Doe"


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    username_claim=["user", "username"],
    groups_claim=["user", "groups.names"],
    sync_groups=True,
)
def test_create_user_with_mapped_groups(dummy_config, callback_request: HttpRequest):
    assert not User.objects.exists()
    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "user": {
                "username": "admin",
                "groups.names": ["supergroup"],
            },
            "email": "admin@localhost",
            "given_name": "John",
            "family_name": "Doe",
        }
    )

    user = backend.authenticate(request=callback_request)

    assert user is not None
    assert isinstance(user, User)

    group_names = set(user.groups.values_list("name", flat=True))
    assert group_names == {"supergroup"}


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    username_claim=["sub"],
    groups_claim=["roles"],
    sync_groups=True,
)
def test_groups_claim_string_instead_of_list(
    dummy_config, callback_request: HttpRequest
):
    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "sub": "admin",
            "email": "admin@localhost",
            "given_name": "John",
            "family_name": "Doe",
            "roles": "admins",
        }
    )

    user = backend.authenticate(request=callback_request)

    assert user is not None
    assert isinstance(user, User)

    assert set(user.groups.values_list("name", flat=True)) == {"admins"}


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    username_claim=["user", "username"],
    groups_claim=["user", "groups.names"],
    claim_mapping={
        "email": ["attributes.email"],
        "first_name": ["profile", "given_name"],
        "last_name": ["profile", "family_name"],
    },
    sync_groups=True,
    sync_groups_glob_pattern="*",
)
def test_update_user_with_custom_and_complex_config(
    dummy_config, callback_request: HttpRequest
):
    # explicitly clear to wipe any possible implicit defaults
    dummy_config.default_groups.clear()
    existing_user = User.objects.create(
        username="admin",
        email="outdated@example.com",
        first_name="O.",
        last_name="Utdated",
        is_superuser=True,
    )
    group = Group.objects.create(name="to-remove")
    existing_user.groups.add(group)

    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "user": {
                "username": "admin",
                "groups.names": ["supergroup"],
            },
            "profile": {
                "given_name": "John",
                "family_name": "Doe",
            },
            "attributes.email": "admin@localhost",
        }
    )

    user = backend.authenticate(request=callback_request)

    assert isinstance(user, User)

    # check that the existing user was correctly found based on the username claim
    assert user == existing_user

    # Verify that a user is created with the correct values
    assert user.username == "admin"
    assert user.email == "admin@localhost"
    assert user.first_name == "John"
    assert user.last_name == "Doe"
    # is_superuser may not be affected if `superuser_group_names` is not set
    assert user.is_superuser

    # missing groups must be created because of sync_groups and sync_groups_glob_pattern
    assert set(user.groups.values_list("name", flat=True)) == {"supergroup"}


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    groups_claim=[],
    sync_groups=True,
    sync_groups_glob_pattern="*",
)
def test_authenticate_user_no_group_sync_without_claim(
    dummy_config, callback_request: HttpRequest
):
    Group.objects.create(name="group1")
    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "sub": "123456",
            "roles": ["group1", "newgroup"],
        }
    )

    user = backend.authenticate(request=callback_request)

    assert isinstance(user, User)
    # Verify that no groups were created or assigned
    assert Group.objects.count() == 1
    assert not user.groups.exists()


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    groups_claim=["roles"],
    sync_groups=True,
    sync_groups_glob_pattern="myapp:*",
)
def test_authenticate_user_groups_glob_pattern(
    dummy_config, callback_request: HttpRequest
):
    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "sub": "123456",
            "roles": ["admin", "myapp:editor"],
        }
    )

    user = backend.authenticate(request=callback_request)

    assert isinstance(user, User)
    # Verify that no groups were created or assigned
    assert Group.objects.count() == 1
    assert set(user.groups.values_list("name", flat=True)) == {"myapp:editor"}


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    groups_claim=["roles"],
    sync_groups=True,
    sync_groups_glob_pattern="myapp:*",
)
def test_authenticate_user_groups_and_default_groups(
    dummy_config, callback_request: HttpRequest
):
    default_group = Group.objects.create(name="default1")
    Group.objects.create(name="default2")
    dummy_config.default_groups.set({default_group})
    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "sub": "123456",
            "roles": ["admin", "myapp:editor"],
        }
    )

    user = backend.authenticate(request=callback_request)

    assert isinstance(user, User)
    # Verify that no groups were created or assigned
    assert Group.objects.count() == 3
    assert set(user.groups.values_list("name", flat=True)) == {
        "myapp:editor",
        "default1",
    }


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    make_users_staff=True,
    claim_mapping={
        "is_superuser": ["is_god"],
    },
)
def test_authenticate_user_make_staff(dummy_config, callback_request: HttpRequest):
    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "sub": "123456",
            "is_god": 1,
        }
    )

    user = backend.authenticate(request=callback_request)

    assert isinstance(user, User)
    assert user.is_staff
    assert user.is_superuser


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    make_users_staff=True,
    sync_groups=False,
    groups_claim=["roles"],
    superuser_group_names=["superuser"],
)
def test_authenticate_user_make_superuser_based_on_group(
    dummy_config, callback_request: HttpRequest
):
    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "sub": "123456",
            "roles": ["superuser", "groupadmin"],
        }
    )

    user = backend.authenticate(request=callback_request)

    assert isinstance(user, User)
    assert user.is_superuser
    assert not Group.objects.exists()


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    make_users_staff=True,
    sync_groups=False,
    groups_claim=["roles"],
    superuser_group_names=["superuser"],
)
def test_remove_superuser_based_on_group(dummy_config, callback_request: HttpRequest):
    existing_user = User.objects.create_user(username="123456", is_superuser=True)
    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "sub": "123456",
            "roles": ["nosuperuser", "groupadmin"],
        }
    )

    user = backend.authenticate(request=callback_request)

    assert isinstance(user, User)
    assert user == existing_user
    assert not user.is_superuser


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    make_users_staff=True,
    sync_groups=False,
    groups_claim=["roles"],
    superuser_group_names=[],
)
def test_do_nothing_if_no_superuser_groups_configured(
    dummy_config, callback_request: HttpRequest
):
    existing_user = User.objects.create_user(username="123456", is_superuser=True)
    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "sub": "123456",
            "roles": ["nosuperuser", "groupadmin"],
        }
    )

    user = backend.authenticate(request=callback_request)

    assert isinstance(user, User)
    assert user == existing_user
    assert user.is_superuser


#
# AUTHENTICATE FLOW PROBLEMS
#


@pytest.mark.oidcconfig(enabled=True)
def test_authenticate_called_without_args(dummy_config):
    backend = OIDCAuthenticationBackend()

    user = backend.authenticate(request=None)

    assert user is None


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    username_claim=["sub"],
)
def test_username_claim_empty(dummy_config, callback_request: HttpRequest):
    backend = MockBackend(claims={"sub": ""})

    user = backend.authenticate(request=callback_request)

    assert user is None


@pytest.mark.oidcconfig(
    enabled=True, userinfo_claims_source=UserInformationClaimsSources.id_token
)
def test_empty_claims(dummy_config, callback_request: HttpRequest):
    backend = MockBackend(claims={})

    user = backend.authenticate(request=callback_request)

    assert user is None


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    username_claim=["sub"],
    groups_claim=["roles"],
    sync_groups=True,
)
def test_groups_claim_wrong_type(dummy_config, callback_request: HttpRequest):
    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "sub": "admin",
            "email": "admin@localhost",
            "given_name": "John",
            "family_name": "Doe",
            "roles": ["group1", None, 123],
        }
    )

    user = backend.authenticate(request=callback_request)

    assert user is not None
    assert isinstance(user, User)

    # fails silently
    assert not user.groups.exists()


def test_authenticate_without_previous_state(rf: RequestFactory):
    request = rf.get("/oidc/callback", {"state": "foo", "code": "bar"})
    backend = OIDCAuthenticationBackend()

    with pytest.raises(BadRequest):
        backend.authenticate(request=request)


#
# CACHE/QUERYING FOR CONFIG DURING PERMISSON CHECK.
#


def test_init_does_not_perform_config_io(mocker):
    """
    Regression test for https://github.com/maykinmedia/mozilla-django-oidc-db/issues/30

    This test will fail if IO is performed in one of two ways:

    * assertion fails because get_solo was called
    * pytest will complain about database access which is forbidden because there is
      no pytest.mark.django_db present (deliberately)
    """
    m_get_setting = mocker.patch(
        "mozilla_django_oidc_db.backends.get_setting_from_config"
    )

    # instantiate
    OIDCAuthenticationBackend()

    m_get_setting.assert_not_called()



from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _

import pytest

from mozilla_django_oidc_db.fields import ClaimFieldDefault
from mozilla_django_oidc_db.models import OpenIDConnectConfig


def test_default_cache_key():
    assert (
        OpenIDConnectConfig.get_cache_key()
        == "solo:mozilla_django_oidc_db:openidconnectconfig"
    )


def test_different_default_provided(settings):
    settings.MOZILLA_DJANGO_OIDC_DB_PREFIX = "otherprefix"
    assert (
        OpenIDConnectConfig.get_cache_key()
        == "otherprefix:mozilla_django_oidc_db:openidconnectconfig"
    )


@pytest.fixture()
def unset_oidc_cache_prefix(settings):
    del settings.MOZILLA_DJANGO_OIDC_DB_PREFIX


def test_validate_claim_mapping_fields():
    instance = OpenIDConnectConfig(
        claim_mapping={
            "bad_field_no_cookie": ["har"],
        }
    )

    with pytest.raises(ValidationError) as exc_context:
        instance.clean()

    err_dict = exc_context.value.message_dict
    assert "claim_mapping" in err_dict
    error = _("Field '{field}' does not exist on the user model").format(
        field="bad_field_no_cookie"
    )
    assert error in err_dict["claim_mapping"]


def test_validate_username_field_not_in_claim_mapping():
    instance = OpenIDConnectConfig(
        claim_mapping={
            "username": ["nope"],
        }
    )

    with pytest.raises(ValidationError) as exc_context:
        instance.clean()

    err_dict = exc_context.value.message_dict
    assert "claim_mapping" in err_dict
    error = _("The username field may not be in the claim mapping")
    assert error in err_dict["claim_mapping"]


def test_claim_field_default_equality():
    assert ClaimFieldDefault("foo", "bar") == ClaimFieldDefault("foo", "bar")
    assert ClaimFieldDefault("foo", "bar") != ClaimFieldDefault("bar", "foo")
